from typing import List
from itertools import chain
from iobes import TokenFunction
from iobes.utils import extract_type, extract_function, replace_prefix


def iob_to_bio(tags: List[str]) -> List[str]:
    new_tags = []
    prev_type = TokenFunction.OUTSIDE
    for token in tags:
        func = extract_function(token)
        _type = extract_type(token)
        # `I-` tags at the beginning need to be converted
        if func == TokenFunction.INSIDE:
            # If we are after an `O` we are the start of an entity and need to switch to `B-`
            # If we are after an `I-` or `B-` of a different type we are a new entity
            if prev_type == TokenFunction.OUTSIDE or _type != prev_type:
                token = f"{TokenFunction.BEGIN}-{_type}"
            # If we are after one of the same type then we are part of that entity and don't change
        # `B-` tags are passed through as is
        new_tags.append(token)
        prev_type = _type
    return new_tags


def iob_to_iobes(tags: List[str]) -> List[str]:
    return bio_to_iobes(iob_to_bio(tags))


def iob_to_bilou(tags: List[str]) -> List[str]:
    return iobes_to_bilou(iob_to_iobes(tags))


def iob_to_bmeow(tags: List[str]) -> List[str]:
    return iobes_to_bmeow(iob_to_iobes(tags))


iob_to_bmewo = iob_to_bmeow


def bio_to_iob(tags: List[str]) -> List[str]:
    new_tags = []
    prev_type = TokenFunction.OUTSIDE
    for token in tags:
        _type = extract_type(token)
        func = extract_function(token)
        # We want to keep `B-`s that represent the transition between two entities of the same type
        # and convert the rest of them to `I-`s
        if func == TokenFunction.BEGIN:
            if prev_type != _type:
                token = f"{TokenFunction.INSIDE}-{_type}"
        # `I-` tags are passed through
        new_tags.append(token)
        prev_type = _type
    return new_tags


def bio_to_iobes(tags: List[str]) -> List[str]:
    new_tags = []
    for c, n in zip(tags, chain(tags[1:], [TokenFunction.OUTSIDE])):
        curr_func = extract_function(c)
        curr_type = extract_type(c)
        next_func = extract_function(n)
        next_type = extract_type(n)
        if curr_func == TokenFunction.BEGIN:
            if next_func == TokenFunction.INSIDE and next_type == curr_type:
                token = c
            else:
                token = f"{TokenFunction.SINGLE}-{curr_type}"
        elif curr_func == TokenFunction.INSIDE:
            if next_func == TokenFunction.INSIDE and next_type == curr_type:
                token = c
            else:
                token = f"{TokenFunction.END}-{curr_type}"
        else:
            token = c
        new_tags.append(token)
    return new_tags


def bio_to_bilou(tags: List[str]) -> List[str]:
    return iobes_to_bilou(bio_to_iobes(tags))


def bio_to_bmeow(tags: List[str]) -> List[str]:
    return iobes_to_bmeow(bio_to_iobes(tags))


bio_to_bmewo = bio_to_bmeow


def iobes_to_iob(tags: List[str]) -> List[str]:
    return bio_to_iob(iobes_to_bio(tags))


def iobes_to_bio(tags: List[str]) -> List[str]:
    return list(
        map(
            lambda x: replace_prefix(
                replace_prefix(x, TokenFunction.END, TokenFunction.INSIDE), TokenFunction.SINGLE, TokenFunction.BEGIN
            ),
            tags,
        )
    )


def iobes_to_bilou(tags: List[str]) -> List[str]:
    return [iobes_to_bilou_token(t) for t in tags]


def iobes_to_bmeow(tags: List[str]) -> List[str]:
    return [iobes_to_bmeow_token(s) for s in tags]


iobes_to_bmewo = iobes_to_bmeow


def bilou_to_iob(tags: List[str]) -> List[str]:
    return iobes_to_iob(bilou_to_iobes(tags))


def bilou_to_bio(tags: List[str]) -> List[str]:
    return iobes_to_bio(bilou_to_iobes(tags))


def bilou_to_iobes(tags: List[str]) -> List[str]:
    return [bilou_to_iobes_token(t) for t in tags]


def bilou_to_bmeow(tags: List[str]) -> List[str]:
    return iobes_to_bmeow(bilou_to_iobes(tags))


bilou_to_bmewo = bilou_to_bmeow


def bmeow_to_iob(tags: List[str]) -> List[str]:
    return iobes_to_iob(bmeow_to_iobes(tags))


bmewo_to_iob = bmeow_to_iob


def bmeow_to_bio(tags: List[str]) -> List[str]:
    return iobes_to_bio(bmeow_to_iobes(tags))


bmewo_to_bio = bmeow_to_bio


def bmeow_to_iobes(tags: List[str]) -> List[str]:
    return [bmeow_to_iobes_token(t) for t in tags]


bmewo_to_iobes = bmeow_to_iobes


def bmeow_to_bilou(tags: List[str]) -> List[str]:
    return iobes_to_bilou(bmeow_to_iobes(tags))


bmewo_to_bilou = bmeow_to_bilou


def bilou_to_iobes_token(token: str) -> str:
    func = extract_function(token)
    _type = extract_type(token)
    if func == TokenFunction.LAST:
        return f"{TokenFunction.END}-{_type}"
    if func == TokenFunction.UNIT:
        return f"{TokenFunction.SINGLE}-{_type}"
    return token


def iobes_to_bilou_token(token: str) -> str:
    func = extract_function(token)
    _type = extract_type(token)
    if func == TokenFunction.END:
        return f"{TokenFunction.LAST}-{_type}"
    if func == TokenFunction.SINGLE:
        return f"{TokenFunction.UNIT}-{_type}"
    return token


def iobes_to_bmeow_token(token: str) -> str:
    func = extract_function(token)
    _type = extract_type(token)
    if func == TokenFunction.INSIDE:
        return f"{TokenFunction.MIDDLE}-{_type}"
    if func == TokenFunction.SINGLE:
        return f"{TokenFunction.WHOLE}-{_type}"
    return token


iobes_to_bmewo_token = iobes_to_bmeow


def bmeow_to_iobes_token(token: str) -> str:
    func = extract_function(token)
    _type = extract_type(token)
    if func == TokenFunction.MIDDLE:
        return f"{TokenFunction.INSIDE}-{_type}"
    if func == TokenFunction.WHOLE:
        return f"{TokenFunction.SINGLE}-{_type}"
    return token


bmewo_to_iobes_token = bmeow_to_iobes_token


def bilou_to_bmeow_token(token: str) -> str:
    return iobes_to_bmeow_token(bilou_to_iobes_token(token))


bilou_to_bmewo_token = bilou_to_bmeow_token


def bmeow_to_bilou_token(token: str) -> str:
    return iobes_to_bilou_token(bmwow_to_iobes_token(token))


bmewo_to_bilou_token = bmeow_to_bilou_token



from itertools import chain, permutations
from typing import List, Tuple, NamedTuple, Dict
from iobes import SpanEncoding, TokenFunction
from iobes.utils import extract_function, extract_type
from iobes.convert import (
    bilou_to_iobes,
    iobes_to_bilou_token,
    bmeow_to_iobes,
    iobes_to_bmeow_token,
)


class Transition(NamedTuple):
    source: str
    target: str
    valid: bool


def transitions_to_tuple_map(transitions: List[Transition]) -> Dict[Tuple[str, str], bool]:
    return {(src, tgt): valid for src, tgt, valid in transitions}


def transitions_to_map(transitions: List[Transition]) -> Dict[str, Dict[str, bool]]:
    mapping = {}
    for src, tgt, valid in transitions:
        to = mapping.setdefault(src, {})
        to[tgt] = valid
    return mapping


def transitions_legality(
    tokens: List[str], span_type: SpanEncoding, start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    if span_type is SpanEncoding.IOB:
        return iob_transitions(tokens, start, end)
    if span_type is SpanEncoding.BIO:
        return bio_transitions(tokens, start, end)
    if span_type is SpanEncoding.IOBES:
        return iobes_transitions(tokens, start, end)
    if span_type is SpanEncoding.BILOU:
        return bilou_transitions(tokens, start, end)
    if span_type is SpanEncoding.BMEOW or span_type is SpanEncoding.BMEWO:
        return bmeow_transitions(tokens, start, end)
    if span_type is SpanEncoding.TOKEN:
        return token_transitions(tokens, start, end)
    raise ValueError(f"Unknown SpanEncoding Scheme, got: `{span_type}`")


def token_transitions_legality(tokens: List[str], start: str, end: str) -> List[Transition]:
    transitions = []
    for src, tgt in permutations(tokens, 2):
        transitions.append(Transition(src, tgt, True))
    for token in tokens:
        transitions.append(Transition(token, token, True))
        transitions.append(Transition(start, token, True))
        transitions.append(Transition(token, start, False))
        transitions.append(Transition(token, end, True))
        transitions.append(Transition(end, token, False))
    return transitions


def iob_transitions_legality(tokens: List[str], start: str, end: str) -> List[Transition]:
    transitions = []
    for src in chain(tokens, [start, end]):
        src_func = extract_function(src)
        src_type = extract_type(src)
        for tgt in chain(tokens, [start, end]):
            tgt_func = extract_function(tgt)
            tgt_type = extract_type(tgt)
            # Can't transition to start
            if tgt == start:
                transitions.append(Transition(src, tgt, False))
                continue
            # Can't transition from start
            if src == end:
                transitions.append(Transition(src, tgt, False))
                continue
            # Can't transition from start to B because B needs to be between two spans of the same type
            elif src == start:
                if tgt_func == TokenFunction.BEGIN:
                    transitions.append(Transition(src, tgt, False))
                    continue
            elif src_func == TokenFunction.BEGIN:
                # Can only go from B to B of the same type
                if tgt_func == TokenFunction.BEGIN:
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == TokenFunction.INSIDE:
                # Can only go from I to B of the same type
                if tgt_func == TokenFunction.BEGIN:
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == TokenFunction.OUTSIDE:
                # Can't start a span with B unless preceded by another span
                if tgt_func == TokenFunction.BEGIN:
                    transitions.append(Transition(src, tgt, False))
                    continue
            transitions.append(Transition(src, tgt, True))
    return transitions


def bio_transitions_legality(tokens: List[str], start: str, end: str) -> List[Transition]:
    transitions = []
    for src in chain(tokens, [start, end]):
        src_func = extract_function(src)
        src_type = extract_type(src)
        for tgt in chain(tokens, [start, end]):
            tgt_func = extract_function(tgt)
            tgt_type = extract_type(tgt)
            # Can't transition to start
            if tgt == start:
                transitions.append(Transition(src, tgt, False))
                continue
            # Can't transition from end
            if src == end:
                transitions.append(Transition(src, tgt, False))
                continue
            elif src == start:
                # Can't go from start to an I
                if tgt_func == TokenFunction.INSIDE:
                    transitions.append(Transition(src, tgt, False))
                    continue
            elif src_func == TokenFunction.BEGIN:
                # Can only go from B to I of same type
                if tgt_func == TokenFunction.INSIDE:
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == TokenFunction.INSIDE:
                # Can only go from I to I of same type
                if tgt_func == TokenFunction.INSIDE:
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == TokenFunction.OUTSIDE:
                # Can't start an entity with I
                if tgt_func == TokenFunction.INSIDE:
                    transitions.append(Transition(src, tgt, False))
                    continue
            transitions.append(Transition(src, tgt, True))
    return transitions


def iobes_transitions_legality(tokens: List[str], start: str, end: str) -> List[Transition]:
    transitions = []
    for src in chain(tokens, [start, end]):
        src_func = extract_function(src)
        src_type = extract_type(src)
        for tgt in chain(tokens, [start, end]):
            tgt_func = extract_function(tgt)
            tgt_type = extract_type(tgt)
            # Can't transition to start
            if tgt == start:
                transitions.append(Transition(src, tgt, False))
                continue
            # Can't transition from end
            if src == end:
                transitions.append(Transition(src, tgt, False))
                continue
            elif src == start:
                # Can't start span with I or E
                if tgt_func in (TokenFunction.INSIDE, TokenFunction.END):
                    transitions.append(Transition(src, tgt, False))
                    continue
            elif src_func == TokenFunction.BEGIN:
                # Can't go from B to B, S, or O because we didn't close the entity
                if tgt_func in (TokenFunction.BEGIN, TokenFunction.SINGLE, TokenFunction.OUTSIDE) or tgt_func == end:
                    transitions.append(Transition(src, tgt, False))
                    continue
                # Can only go from B to I or E of the same type
                elif tgt_func in (TokenFunction.INSIDE, TokenFunction.END):
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == TokenFunction.INSIDE:
                # Can't from from I to B, S, or O because we didin't close the entity
                if tgt_func in (TokenFunction.BEGIN, TokenFunction.SINGLE, TokenFunction.OUTSIDE) or tgt == end:
                    transitions.append(Transition(src, tgt, False))
                    continue
                # Can only go from I to I or E of the same Type
                elif tgt_func in (TokenFunction.INSIDE, TokenFunction.END):
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func in (TokenFunction.END, TokenFunction.SINGLE, TokenFunction.OUTSIDE):
                # Going from outside an entity (or ending it) to one that was inside the entity (I/E) is illegal
                if tgt_func in (TokenFunction.INSIDE, TokenFunction.END):
                    transitions.append(Transition(src, tgt, False))
                    continue
            # Other transitions are allowed
            transitions.append(Transition(src, tgt, True))
    return transitions


def bilou_transitions_legality(tokens: List[str], start: str, end: str) -> List[Transition]:
    tokens = bilou_to_iobes(tokens)
    transitions = iobes_transitions(tokens, start, end)
    new_trans = []
    for trans in transitions:
        new_trans.append(
            Transition(iobes_to_bilou_token(trans.source), iobes_to_bilou_token(trans.target), trans.valid)
        )
    return new_trans


def bmeow_transitions_legality(tokens: List[str], start: str, end: str) -> List[Transition]:
    tokens = bmeow_to_iobes(tokens)
    transitions = iobes_transitions(tokens, start, end)
    new_trans = []
    for trans in transitions:
        new_trans.append(
            Transition(iobes_to_bmeow_token(trans.source), iobes_to_bmeow_token(trans.target), trans.valid)
        )
    return new_trans


bmewo_transitions_legality = bmeow_transitions_legality



__version__ = "1.2.0"

import logging
from enum import Enum
from typing import NamedTuple, List


LOGGER = logging.getLogger("iobes")


class TokenFunction:
    OUTSIDE = "O"
    BEGIN = "B"
    INSIDE = "I"
    MIDDLE = "M"
    END = "E"
    LAST = "L"
    SINGLE = "S"
    UNIT = "U"
    WHOLE = "W"
    GO = "<GO>"
    EOS = "<EOS>"


class SpanEncoding(Enum):
    IOB = 1
    BIO = 2
    IOBES = 3
    BILOU = 4
    BMEOW = 5
    BMEWO = 6
    TOKEN = 7

    @classmethod
    def from_string(cls, value):
        value = value.lower().strip()
        if value == "iob":
            return cls.IOB
        if value in ("iob2", "bio"):
            return cls.BIO
        if value == "iobes":
            return cls.IOBES
        if value in ("bilou", "bioul"):
            return cls.BILOU
        if value in ("bmewo", "bmeow"):
            return cls.BMEOW
        if value == "token":
            return cls.TOKEN
        raise ValueError(f"Unknown Encoding Scheme, got: `{value}`")


class Span(NamedTuple):
    type: str
    start: int
    end: int
    tokens: List[int]


class Error(NamedTuple):
    location: int
    type: str
    current: str
    previous: str
    next: str


from iobes.convert import (
    iob_to_bio,
    iob_to_iobes,
    iob_to_bilou,
    iob_to_bmeow,
    iob_to_bmewo,
    bio_to_iob,
    bio_to_iobes,
    bio_to_bilou,
    bio_to_bmeow,
    bio_to_bmewo,
    iobes_to_iob,
    iobes_to_bio,
    iobes_to_bilou,
    iobes_to_bmeow,
    iobes_to_bmewo,
    bilou_to_iob,
    bilou_to_bio,
    bilou_to_iobes,
    bilou_to_bmeow,
    bilou_to_bmewo,
    bmeow_to_iob,
    bmeow_to_bio,
    bmeow_to_iobes,
    bmeow_to_bilou,
    bmewo_to_iob,
    bmewo_to_bio,
    bmewo_to_iobes,
    bmewo_to_bilou,
)
from iobes.parse import (
    parse_spans,
    parse_spans_token,
    parse_spans_iob,
    parse_spans_bio,
    parse_spans_iobes,
    parse_spans_bilou,
    parse_spans_bmeow,
    parse_spans_bmewo,
    parse_spans_with_errors,
    parse_spans_token_with_errors,
    parse_spans_iob_with_errors,
    parse_spans_bio_with_errors,
    parse_spans_iobes_with_errors,
    parse_spans_bilou_with_errors,
    parse_spans_bmeow_with_errors,
    parse_spans_bmewo_with_errors,
    validate_labels,
    validate_labels_iob,
    validate_labels_bio,
    validate_labels_iobes,
    validate_labels_bilou,
    validate_labels_bmeow,
    validate_labels_bmewo,
)
from iobes.transition import (
    Transition,
    transitions_legality,
    iob_transitions_legality,
    bio_transitions_legality,
    iobes_transitions_legality,
    bilou_transitions_legality,
    bmeow_transitions_legality,
    bmewo_transitions_legality,
    transitions_to_tuple_map,
    transitions_to_map,
)
from iobes.utils import (
    extract_type,
    extract_function,
)



from typing import List, Union, Any


def extract_type(token: str, sep: str = "-") -> str:
    if sep not in token:
        return token
    return token.split(sep, maxsplit=1)[1]


def extract_function(token: str, sep: str = "-") -> str:
    if sep not in token:
        return token
    return token.split(sep, maxsplit=1)[0]


def replace_prefix(string: str, match: str, sub: str) -> str:
    if string.startswith(match):
        return sub + string[len(match) :]
    return string[:]


def safe_get(xs: List[str], idx: int) -> Union[Any, None]:
    if idx < 0 or idx >= len(xs):
        return None
    return xs[idx]
